"""
Main engine for PyTorch
"""

from __future__ import annotations
from typing import Optional, Union, Callable, Dict
from contextlib import nullcontext

import os
import torch
import torch.utils.data.datapipes as dp
from torch import autocast
from torch.cuda import amp
from torchdata.dataloader2 import DataLoader2
from random import random

from returnn.config import Config
from returnn.log import log
from returnn.engine.base import EngineBase
import returnn.frontend as rf
from returnn.tensor import TensorDict, Tensor
from returnn.datasets.basic import init_dataset, Dataset
from returnn.util import basic as util
from returnn.util import NumbersDict
from returnn.util.basic import NotSpecified
from .updater import Updater
from .data import pipeline as data_pipeline
from .data import returnn_dataset_wrapper
from .frontend.bridge import rf_module_to_pt_module


class Engine(EngineBase):
    """
    PyTorch engine
    """

    def __init__(self, config: Config):
        """
        :param config:
        """
        super(Engine, self).__init__(config=config)
        rf.select_backend_torch()
        self.model_filename = self.config.value("model", None)
        self._mp_manager = torch.multiprocessing.Manager()
        self._epoch_mp_shared = self._mp_manager.Value("i", 0)
        self.train_dataset = None  # type: Optional[Dataset]
        self.eval_datasets = {}
        self.extern_data = None  # type: Optional[TensorDict]
        self._train_dataloader = None  # type: Optional[DataLoader2]
        self._eval_dataloaders = {}  # type: Dict[str, DataLoader2]

        self._start_epoch = None  # type: Optional[int]
        self._final_epoch = None  # type: Optional[int]
        self._orig_model = None  # type: Optional[Union[rf.Module, torch.nn.Module]]
        self._pt_model = None  # type: Optional[torch.nn.Module]
        self._train_step_func = None  # type: Optional[Callable]
        self._save_model_epoch_interval = 1
        self._updater = None  # type: Optional[Updater]

        self._use_autocast = False
        self._autocast_dtype = None  # type: Optional[str]
        self._grad_scaler = None  # type: Optional[amp.GradScaler]

        self._device = _get_device_from_config(config)
        print("Using device:", self._device, file=log.v2)

        amp_options = self.config.typed_value("torch_amp")
        grad_scaler_opts = self.config.typed_value("grad_scaler", NotSpecified)
        if amp_options is not None:
            self._use_autocast = True
            if isinstance(amp_options, dict):
                amp_options = util.CollectionReadCheckCovered(amp_options)
                dtype = amp_options.get("dtype", None)
                grad_scaler_opts = amp_options.get("grad_scaler", grad_scaler_opts)
                amp_options.assert_all_read()
            elif isinstance(amp_options, str):
                dtype = amp_options
            else:
                raise TypeError(f"Invalid type for torch_amp: {type(amp_options)}")
            if isinstance(dtype, str):
                dtype = getattr(torch, dtype)
            assert isinstance(dtype, torch.dtype) or dtype is None
            print(f"Using autocast (automatic mixed precision (AMP)) with dtype {dtype}", file=log.v2)
            self._autocast_dtype = dtype

        if grad_scaler_opts is NotSpecified:
            grad_scaler_opts = {} if self._use_autocast else None
        if grad_scaler_opts is not None:
            assert isinstance(grad_scaler_opts, dict)
            print("Using GradScaler with options:", grad_scaler_opts, file=log.v2)
            self._grad_scaler = amp.GradScaler(**grad_scaler_opts)

    def init_train_from_config(
        self,
        config: Optional[Config] = None,
        train_data: Optional[Dataset] = None,
        dev_data: Optional[Dataset] = None,
        eval_data: Optional[Dataset] = None,
    ):
        """
        :param config:
        :param train_data:
        :param dev_data:
        :param eval_data:
        """
        assert config is self.config
        super().init_train_from_config(config=config)
        self.train_dataset = train_data
        self.eval_datasets.clear()
        if dev_data:
            self.eval_datasets["dev"] = dev_data
        if eval_data:
            self.eval_datasets["eval"] = eval_data
        if config.has("eval_datasets"):
            for dataset_name, dataset_opts in config.typed_value("eval_datasets", {}).items():
                self.eval_datasets[dataset_name] = init_dataset(dataset_opts, default_kwargs={"name": dataset_name})

        extern_data = TensorDict()
        extern_data_dict = self.config.typed_value("extern_data")
        extern_data.update(extern_data_dict, auto_convert=True)
        self.extern_data = extern_data

        self._train_dataloader = self._create_data_loader(train_data) if train_data else None
        for dataset_name, dataset in self.eval_datasets.items():
            self._eval_dataloaders[dataset_name] = self._create_data_loader(dataset)

        self._start_epoch = self.get_train_start_epoch(self.config)
        self._final_epoch = self.config_get_final_epoch(self.config)

        self._load_model(epoch=self._start_epoch)
        self._save_model_epoch_interval = config.int("save_interval", 1)

        self._updater = Updater(self.config, self._pt_model, self.learning_rate)
        self._updater.create_optimizer()
        if self._start_epoch > 1:
            self._load_optimizer(self._start_epoch)

        self._train_step_func = self.config.typed_value("train_step")
        assert self._train_step_func, "train_step not defined"

    def train(self):
        """
        Main training loop.
        """

        print("Starting training at epoch {}.".format(self._start_epoch), file=log.v3)
        assert self._pt_model is not None, "Model not initialized, call init_train_from_config()."

        self.epoch = self._start_epoch
        self._epoch_mp_shared.value = self.epoch
        while self.epoch <= self._final_epoch:
            self.init_train_epoch()
            self.train_epoch()

            self.epoch += 1
            self._epoch_mp_shared.value = self.epoch

        print("Finished training at epoch {}.".format(self.epoch), file=log.v3)

    def init_train_epoch(self):
        """
        init train (sub)epoch. LR etc
        """
        self.learning_rate = self.learning_rate_control.get_learning_rate_for_epoch(self.epoch)

        # Update learning rate
        self._updater.set_learning_rate(self.learning_rate)

    def train_epoch(self):
        """
        train one (sub)epoch
        """
        print("start", self.get_epoch_str(), "with learning rate", self.learning_rate, "...", file=log.v4)

        self._pt_model.train()

        accumulated_losses_dict = NumbersDict()
        accumulated_inv_norm_factors_dict = NumbersDict()
        step_idx = 0
        for data in self._train_dataloader:
            self._updater.get_optimizer().zero_grad()
            self._run_step(data, train_flag=True)

            train_ctx = rf.get_run_ctx()
            total_loss = train_ctx.total_loss()
            losses_dict = NumbersDict(
                {
                    name: float(loss.get_summed_loss().raw_tensor.detach().cpu().numpy())
                    for name, loss in train_ctx.losses.items()
                }
            )
            inv_norm_factors_dict = NumbersDict(
                {name: float(_to_raw(loss.get_inv_norm_factor())) for name, loss in train_ctx.losses.items()}
            )

            if self._grad_scaler is not None:
                self._grad_scaler.scale(total_loss).backward()
                self._grad_scaler.step(self._updater.get_optimizer())
                self._grad_scaler.update()
            else:
                total_loss.raw_tensor.backward()
                self._updater.get_optimizer().step()

            accumulated_losses_dict += losses_dict
            accumulated_inv_norm_factors_dict += inv_norm_factors_dict
            _print_process(
                f"ep {self.epoch} train",
                step=step_idx,
                eval_info=dict(losses_dict / inv_norm_factors_dict),
            )

            step_idx += 1
            self.global_train_step += 1

        print("Trained %i steps" % step_idx)

        accumulated_losses_dict = accumulated_losses_dict / accumulated_inv_norm_factors_dict
        self.learning_rate_control.set_epoch_error(
            self.epoch, {f"train_loss_{k}": v for k, v in accumulated_losses_dict.items()}
        )
        self.learning_rate_control.save()

        print(f"Total train loss:", _format_score(dict(accumulated_losses_dict)), file=log.v3)

        if self.epoch % self._save_model_epoch_interval == 0 or self.epoch == self._final_epoch:
            self._save_model()
            self._save_optimizer()

        self.eval_model()

    def eval_model(self):
        """
        Runs model on all eval datasets and calculates the loss.
        """
        self._pt_model.eval()

        eval_dump_str = []
        score_keys = None
        error_keys = None

        for dataset_name, dataset in self.eval_datasets.items():
            print(f"Evaluating dataset {dataset_name!r}", file=log.v3)

            data_loader = self._eval_dataloaders[dataset_name]

            accumulated_losses_dict = NumbersDict()
            accumulated_inv_norm_factors_dict = NumbersDict()
            step_idx = 0

            with torch.no_grad():
                for data in data_loader:

                    self._run_step(data)
                    train_ctx = rf.get_run_ctx()

                    if score_keys is None:
                        score_keys = [name for name, loss in train_ctx.losses.items() if not loss.as_error]
                        error_keys = [name for name, loss in train_ctx.losses.items() if loss.as_error]

                    losses_dict = NumbersDict(
                        {
                            name: float(loss.get_summed_loss().raw_tensor.detach().cpu().numpy())
                            for name, loss in train_ctx.losses.items()
                        }
                    )
                    inv_norm_factors_dict = NumbersDict(
                        {name: float(_to_raw(loss.get_inv_norm_factor())) for name, loss in train_ctx.losses.items()}
                    )

                    accumulated_losses_dict += losses_dict
                    accumulated_inv_norm_factors_dict += inv_norm_factors_dict
                    _print_process(
                        f"ep {self.epoch} {dataset_name} eval",
                        step=step_idx,
                        eval_info=dict(losses_dict / inv_norm_factors_dict),
                    )
                    step_idx += 1

            assert step_idx > 0, f"No data in dataset {dataset_name!r}."
            accumulated_losses_dict = accumulated_losses_dict / accumulated_inv_norm_factors_dict

            self.learning_rate_control.set_epoch_error(
                self.epoch, {f"{dataset_name}_loss_{k}": v for k, v in accumulated_losses_dict.items()}
            )
            self.learning_rate_control.save()

            # Same format as the TF engine.
            eval_dump_str += [
                "%s: score %s error %s"
                % (
                    dataset_name,
                    _format_score({name: accumulated_losses_dict[name] for name in score_keys}),
                    _format_score({name: accumulated_losses_dict[name] for name in error_keys}),
                )
            ]

        print(" ".join(eval_dump_str), file=log.v1)

    def _create_data_loader(self, dataset: Dataset) -> DataLoader2:
        """
        :param dataset: RETURNN dataset
        :return: PyTorch data loader created from given RETURNN dataset
        """
        # Make sure that _dataset_reset does not keep a ref to `self`,
        # otherwise it would trigger to pickle `self` and all its members.
        dataset_reset = returnn_dataset_wrapper.ReturnnDatasetResetMpSharedEpochCallback(
            dataset=dataset, epoch_mp_shared=self._epoch_mp_shared
        )

        wrapped_dataset = returnn_dataset_wrapper.ReturnnDatasetIterDataPipe(dataset, reset_callback=dataset_reset)

        chunking = self.config.typed_value("chunking", None)
        if chunking:
            wrapped_dataset = data_pipeline.ChunkingIterDataPipe(wrapped_dataset, chunking)

        batch_size = self.config.typed_value("batch_size", 1)
        max_seqs = self.config.int("max_seqs", -1)
        batches_dataset = data_pipeline.BatchingIterDataPipe(wrapped_dataset, batch_size=batch_size, max_seqs=max_seqs)
        batches_dataset = dp.iter.Collator(batches_dataset, collate_fn=data_pipeline.collate_batch)

        try:
            return DataLoader2(batches_dataset)
        except TypeError as exc:
            try:
                # noinspection PyPackageRequirements
                import dill
            except ImportError:
                raise ModuleNotFoundError("Possible type error in DataLoader2 due to missing module 'dill'") from exc
            raise

    def _run_step(self, extern_data_raw: Dict[str, torch.Tensor], *, train_flag: bool = False):
        """
        :param dict[str, torch.Tensor] extern_data_raw: model inputs for the step
        """
        assert isinstance(extern_data_raw, dict) and extern_data_raw
        batch_dim = next(iter(self.extern_data.data.values())).dims[0]
        batch_dim.dyn_size_ext = None  # if it is dynamic, reset now, and set it below
        extern_data = TensorDict()
        for k, data in self.extern_data.data.items():
            data = data.copy_template()
            raw_tensor = extern_data_raw[k].to(self._device)
            data.dtype = str(raw_tensor.dtype).split(".")[-1]  # just overwrite for now...
            data.raw_tensor = raw_tensor

            if batch_dim.size is None and batch_dim.dyn_size_ext is None:
                batch_dim.dyn_size_ext = Tensor(batch_dim.name or "batch", dims=[], dtype="int32")
                batch_dim.dyn_size_ext.raw_tensor = torch.tensor(extern_data_raw[k].shape[0], dtype=torch.int32)

            if k + ":seq_len" in extern_data_raw:
                # Sequence lengths have to be on CPU for the later call to rnn.pack_padded_sequence
                size = extern_data_raw[k + ":seq_len"].cpu()
                size_dtype = str(size.dtype).split(".")[-1]
                if data.dims[1].dyn_size_ext is None:
                    data.dims[1].dyn_size_ext = Tensor(data.dims[1].name or "time", dims=[batch_dim], dtype=size_dtype)
                data.dims[1].dyn_size_ext.dtype = size_dtype
                data.dims[1].dyn_size_ext.raw_tensor = size

            extern_data.data[k] = data

        rf.init_train_step_run_ctx(train_flag=train_flag)

        with autocast(device_type=self._device, dtype=self._autocast_dtype) if self._use_autocast else nullcontext():
            sentinel_kw = {"__fwd_compatible_random_arg_%i" % int(random() * 100): None}
            self._train_step_func(model=self._orig_model, extern_data=extern_data, **sentinel_kw)

    def _load_model(self, *, epoch: int):
        """
        Sets self._model to a torch.nn.Module.

        :param epoch:
        """
        checkpoint_state = None
        if epoch > 1:
            filename = self.get_epoch_model_filename(epoch=epoch - 1) + util.get_model_filename_postfix()
            print("Load model %s" % (filename,), file=log.v4)
            checkpoint_state = torch.load(filename)
            assert checkpoint_state["epoch"] == epoch - 1
            step = checkpoint_state["step"]
        else:
            step = 0
        self.global_train_step = step

        # See :mod:`rf.rand` docstring for an explanation of this logic.
        random_seed = self.config.int("random_seed", 42)
        random_seed = (epoch * 193939 + step * 19937 + random_seed * 27644437 + 479001599) % (2**31)
        rf.set_random_seed(random_seed)

        get_model_func = self.config.typed_value("get_model")
        assert get_model_func, "get_model not defined"
        sentinel_kw = {"__fwd_compatible_random_arg_%i" % int(random() * 100): None}
        model = get_model_func(epoch=epoch, step=step, **sentinel_kw)
        self._orig_model = model
        if isinstance(model, rf.Module):
            self._pt_model = rf_module_to_pt_module(model)
        elif isinstance(model, torch.nn.Module):
            self._pt_model = model
        else:
            raise TypeError(f"get_model returned {model} of type {type(model)}, expected rf.Module or torch.nn.Module")
        assert isinstance(self._pt_model, torch.nn.Module)
        print("Model:", self._pt_model, file=log.v4)

        if checkpoint_state is not None:
            self._pt_model.load_state_dict(checkpoint_state["model"])
        preload_from_files = self.config.typed_value("preload_from_files", {})
        if preload_from_files:
            # see `preload_from_files` in tf engine and `returnn.tf.network.CustomCheckpointLoader`
            is_training = self.config.value("task", "train") == "train"
            is_first_train_epoch = epoch == 1 and (
                is_training or self.config.value("task", "train") == "initialize_model"
            )
            # We use the reversed sorted order here to achieve consistent behavior with the TF engine.
            # There, the keys are used in sorted order but if a variable is loaded,
            # it will not be considered anymore afterward.
            # So the first occurrence is used.
            # Here, we overwrite variables even if they have been loaded before.
            # In order to get consistent behavior, we use the reversed order.
            for preload_key, opts in reversed(sorted(preload_from_files.items())):
                assert isinstance(opts, dict) and "filename" in opts
                if opts.get("init_for_train", False):
                    if not is_first_train_epoch:
                        continue
                else:  # default: init for recog
                    if is_training:
                        continue
                print(f"Pre-load weights for key '{preload_key}' from {opts['filename']}", file=log.v3)
                preload_model_state = torch.load(opts["filename"])
                if opts.get("checkpoint_key", "model") is not None:
                    # This can be used if an external checkpoint saves a checkpoint a different structure that just the
                    # model state dict. E.g., if a checkpoint is created using
                    # `torch.save({"model": model.state_dict(), "optimizer": optimizer.state)_dict(), ...})`
                    # we can set checkpoint_key = "model" to load the model.
                    # Currently, this only supports single level dicts, but it could be extended if needed.
                    preload_model_state = preload_model_state[opts.get("checkpoint_key", "model")]
                if opts.get("prefix", ""):
                    # Only params with this prefix should be loaded.
                    # They are expected to be in the checkpoint without this prefix.
                    # By adding the prefix to all params,
                    # we make sure that only params matching this condition are loaded.
                    # This is in line with the behavior of the TF engine.
                    preload_model_state = {opts["prefix"] + key: value for key, value in preload_model_state.items()}
                ignore_params = opts.get("ignore_params", [])
                ignore_params_prefixes = opts.get("ignore_params_prefixes", [])
                for key in list(preload_model_state.keys()):
                    if key in ignore_params or any(
                        [key.startswith(ignore_key) for ignore_key in ignore_params_prefixes]
                    ):
                        print(f"Ignoring variable {key}", file=log.v3)
                        preload_model_state.pop(key)
                for new_name, name_in_checkpoint in opts.get("var_name_mapping", {}).items():
                    preload_model_state[new_name] = preload_model_state.pop(name_in_checkpoint)
                missing_keys, _ = self._pt_model.load_state_dict(preload_model_state, strict=False)
                if missing_keys and not opts.get("ignore_missing", False):
                    prefix_keys = [key for key in self._pt_model.state_dict() if key.startswith(opts.get("prefix", ""))]
                    missing_prefix_keys = set(prefix_keys).intersection(set(missing_keys))
                    assert not missing_prefix_keys, f"Missing keys and ignore_missing=False: {missing_prefix_keys}"
                print(f"Missing keys: {missing_keys}", file=log.v4)

        self._pt_model.to(self._device)

    def _save_model(self):
        """
        Saves the state of self._model to file.
        """
        filename = self.get_epoch_model_filename() + util.get_model_filename_postfix()
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        print("Save model under %s" % (filename,), file=log.v4)
        torch.save(
            {"model": self._pt_model.state_dict(), "epoch": self.epoch, "step": self.global_train_step}, filename
        )

    def _load_optimizer(self, epoch):
        """
        Loads a torch.optim.Optimizer from disk and uses it as the optimizer.
        This function is a wrapper to Updater.load_optimizer().

        :param int epoch: Epoch from which to load the optimizer state.
        """
        filename = self.get_epoch_model_filename(epoch=epoch - 1) + ".opt" + util.get_model_filename_postfix()
        self._updater.load_optimizer(filename)

    def _save_optimizer(self):
        """
        Saves the optimizer state to a file.
        This function is a wrapper to Updater.save_optimizer().
        """
        filename = self.get_epoch_model_filename() + ".opt" + util.get_model_filename_postfix()
        directory = os.path.dirname(filename)
        if directory and not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        self._updater.save_optimizer(filename)

        # keep only the last two optimizer states (two in case one file gets corrupted)
        clean_epoch = self.epoch - 2
        if clean_epoch > 0:
            filename = self.get_epoch_model_filename(epoch=clean_epoch) + ".opt" + util.get_model_filename_postfix()
            if os.path.isfile(filename):
                os.unlink(filename)


def _to_raw(n: Union[int, float, Tensor]):
    if isinstance(n, (int, float)):
        return n
    if isinstance(n, Tensor):
        return n.raw_tensor.detach().cpu().numpy()
    raise TypeError(f"Unexpected {n} of type {type(n)}")


def _print_process(report_prefix, step, eval_info):
    """
    Similar but simplified from TF engine _print_process.

    :param str report_prefix:
    :param int step:
    :param dict[str] eval_info: via :func:`_collect_eval_info`
    :return: nothing, will be printed to log
    """
    if log.verbose[5]:
        info = [report_prefix, "step %i" % step]
        if eval_info:  # Such as score.
            info += ["%s %s" % item for item in sorted(eval_info.items())]
        print(", ".join(filter(None, info)), file=log.v5)


def _format_score(score: Dict[str, float]) -> str:
    """
    Like the TF engine format_score.

    :param score:
    :return: score(s) as str
    """
    if not score:
        return "None"
    if len(score) == 1:
        return str(list(score.values())[0])
    return " ".join(["%s %s" % (key.split(":", 2)[-1], str(score[key])) for key in sorted(score.keys())])


def _get_gpu_device() -> Optional[str]:
    if torch.cuda.is_available():
        return "cuda"
    if hasattr(torch.backends, "mps") and torch.backends.mps.is_available() and torch.backends.mps.is_built():
        return "mps"
    return None


def _get_device_from_config(config: Config) -> str:
    device = config.value("device", None)
    if not device:
        device = _get_gpu_device()
        if device:
            return device
        return "cpu"
    if device == "gpu":
        device = _get_gpu_device()
        if not device:
            raise Exception("No GPU device found, but config requested 'gpu' device.")
    return device



"""
Run context

We can either be in param-init stage,
or in the main training loop,
or forwarding loop.
"""

from __future__ import annotations
from typing import Optional, Union, Any, Sequence, Dict
from dataclasses import dataclass
from returnn.tensor import Tensor, Dim, TensorDict
import returnn.frontend as rf
from . import _backend


__all__ = ["RunCtx", "Loss", "get_run_ctx", "init_train_step_run_ctx", "init_forward_step_run_ctx"]


_run_ctx = None  # type: Optional[RunCtx]
_init_run_ctx = None  # type: Optional[RunCtx]


def reset_run_ctx():
    """
    If we get out of a train step or forward step.
    """
    global _run_ctx
    _run_ctx = None


def init_train_step_run_ctx(*, train_flag: Union[bool, Tensor]):
    """
    Call this at the beginning of a new train step.
    """
    global _run_ctx
    _run_ctx = RunCtx(stage="train_step", train_flag=train_flag)


def init_forward_step_run_ctx(expected_outputs: Optional[TensorDict] = None):
    """
    Call this at the beginning of a new forward step.
    """
    global _run_ctx
    _run_ctx = RunCtx(stage="forward_step", expected_outputs=expected_outputs)


def get_run_ctx() -> RunCtx:
    """
    :return: current run context, see :class:`RunCtx`
    """
    global _run_ctx, _init_run_ctx
    if _run_ctx is None:
        if _init_run_ctx is None:
            _init_run_ctx = RunCtx(stage="init")
        return _init_run_ctx
    return _run_ctx


class RunCtx:
    """
    We can either be in param-init stage,
    or in the main training (or eval) loop,
    or forwarding loop (doing recog, beam search, dumping whatever, ...).

    In training/eval, we expect that some loss is being defined via mark_as_loss().
    In forwarding, we expect that some output is being defined via mark_as_output().
    """

    def __init__(
        self, *, stage: str, train_flag: Union[bool, Tensor] = False, expected_outputs: Optional[TensorDict] = None
    ):
        """
        :param stage:
            - "init"
            - "train_step", also for eval, for mark_as_loss and get_total_loss
            - "forward_step", for mark_as_output
        """
        self._stage = stage
        self._train_flag = train_flag
        self.losses = {}  # type: Dict[str, Loss]
        self.outputs = TensorDict()
        self.expected_outputs = expected_outputs

    @property
    def stage(self) -> str:
        """
        :return: "init", "train_step", "forward_step"
        """
        return self._stage

    @property
    def train_flag(self) -> Union[bool, Tensor]:
        """
        :return: whether we are in training mode, i.e. the model is updated,
            and we are supposed to use dropout and similar mechanisms.
            In a graph-based backend, this can be dynamic.
        """
        return self._train_flag

    def mark_as_loss(
        self,
        loss: Union[Tensor, Any],
        name: str,
        *,
        scale: float = 1.0,
        as_error: bool = False,
        use_normalized_loss: bool = False,
        use_flatten_frames: bool = True,
        custom_inv_norm_factor: Optional[Tensor] = None,
    ) -> None:
        """
        Mark the given loss tensor as a loss.
        This has the effect that it is specially handled by RETURNN.
        Specifically, the optimizer can use it in training,
        and it is used for reporting per batch or per epoch,
        and for learning rate scheduling.

        This currently uses :class:`AsIsLoss` in RETURNN
        but this is an implementation detail and might change.

        :param loss: it should not be reduced - this will be done later, and this is important to properly
            do accumulation taking different seq lengths into account.
            A :class:`Tensor` is usually expected, but a raw tensor is also possible.
        :param name: name of the loss. this name is used for reporting by RETURNN, and also for LR scheduling.
        :param scale: scale the loss by this factor for the training optimizer
          (but not for any reporting). setting to 0.0 has the effect that this loss is not used by the optimizer.
        :param as_error: if True, this loss is reported as an error instead of a loss,
          and not used by the training optimizer.
          This is by convention sth like the frame-error or edit-distance, and usually not differentiable anyway.
        :param bool use_flatten_frames: If True, will use :func:`returnn.tf.util.basic.flatten_with_seq_len_mask`,
          i.e. a "packed" sequence with the padded frames removed, and accumulates over that.
          This can be more efficient, also because it will further optimize incoming computations
          and e.g. skip softmax computations right before on the padded frames.
          This can also avoid issues with inf/nan in some cases.
          If False, it will mask the loss to 0 in the padded frames and accumulate over that.
          Typically, setting this to True (default) is both more efficient and better.
        :param bool use_normalized_loss: the loss used in optimization will be normalized.
          E.g. if the overall normalization is sum(loss)/sum(num_frames), this is also what the optimizer will use,
          otherwise the optimizer will just use sum(loss).
        :param custom_inv_norm_factor:
          The standard inv norm factor is sum(target_seq_len) if the target has a time-axis,
          or sum(output_seq_len) if there is no target and the output has a time-axis,
          or 1 otherwise. (See :func:`Loss.init` for details.)
          This is used for proper normalization of accumulated loss/error per epoch
          and also proper normalization per batch for reporting,
          no matter if use_normalized_loss is True or False.
          If you want to change this norm factor, you can set this.
          Basically, for all reporting, it uses sum(loss) / sum(custom_inv_norm_factor).
        """
        assert self.stage == "train_step"
        if not isinstance(loss, Tensor):
            assert isinstance(loss, _backend.global_backend.RawTensorType)
            loss = rf.convert_to_tensor(loss)
        assert name not in self.losses
        self.losses[name] = Loss(
            loss=loss,
            name=name,
            scale=scale,
            as_error=as_error,
            use_normalized_loss=use_normalized_loss,
            use_flatten_frames=use_flatten_frames,
            custom_inv_norm_factor=custom_inv_norm_factor,
        )

    def mark_as_output(self, tensor: Union[Tensor, Any], name: str, *, dims: Optional[Sequence[int]] = None) -> None:
        """
        Mark this as an output.
        This has the effect that RETURNN will in any case construct the corresponding layer.
        Also see :func:`mark_as_default_output`.

        This is intended mostly for forwarding, or exporting the model (TF graph, TFLite, ONNX, etc).
        You must specify a shape to have the output shape (order of dims) well-defined
        (if not specified, we check if some defaults are possible, like BTF, or BF).

        :param tensor:
        :param name:
        :param dims: this specifies the order of the dims of the output, such that it is well-defined
            for some external application.
            If not specified, we try to infer BTF or BF as default, if that works, otherwise it will be an error.
        """
        assert self.stage == "forward_step"

        if self.expected_outputs is not None:
            assert (
                name in self.expected_outputs.data
            ), f"mark_as_output: unexpected output {name!r}, we expect outputs: {self.expected_outputs}"
        expected_output = self.expected_outputs.data[name] if self.expected_outputs else None
        if dims is None and expected_output:
            dims = expected_output.dims
        if dims is not None and expected_output:
            assert (
                expected_output.dims == dims
            ), f"mark_as_output: {name!r} dims mismatch from expected output, given {dims}, expected {expected_output}"

        if not isinstance(tensor, Tensor):
            assert isinstance(tensor, _backend.global_backend.RawTensorType)
            tensor = rf.convert_to_tensor(tensor, dims=dims)
            # In case it was not specified, just accept whatever order we got.
            dims = tensor.dims

        if dims is None:
            # We try some reasonable defaults, specifically: BTF or BF.
            dims = _default_dim_order(tensor)
        assert set(dims) == set(tensor.dims), f"mark_as_output: tensor {tensor} does not have the dims {dims}"
        tensor = tensor.copy_transpose(dims, allow_int=False)
        tensor = tensor.copy(name=name)
        assert name not in self.outputs.data
        self.outputs.data[name] = tensor

        if expected_output:  # Sanity checks for expected output
            assert expected_output.dims == tensor.dims, (
                f"mark_as_output: {name!r} dims mismatch from expected output,"
                f" given {dims}, expected {expected_output}"
            )
            assert expected_output.dtype == tensor.dtype, (
                f"mark_as_output: {name!r} dtype mismatch from expected output,"
                f" given {tensor.dtype}, expected {expected_output.dtype}"
            )
            assert expected_output.sparse_dim == tensor.sparse_dim, (
                f"mark_as_output: {name!r} sparse_dim mismatch from expected output,"
                f" given {tensor.sparse_dim}, expected {expected_output.sparse_dim}"
            )

    def mark_as_default_output(self, tensor: Union[Tensor, Any], *, shape: Optional[Sequence[Dim]] = None) -> None:
        """
        Calls mark_as_output(tensor, "output", shape=shape).

        Mark this as the default output.
        See :func:`Frontend.mark_as_default_output` for more details.

        :param tensor:
        :param shape:
        """
        self.mark_as_output(tensor, "output", dims=shape)

    def check_outputs_complete(self):
        """
        If expected outputs are given, check that all expected outputs are present.
        """
        if self.expected_outputs:
            assert set(self.expected_outputs.data.keys()) == set(self.outputs.data.keys()), (
                f"check_outputs_complete: expected outputs {self.expected_outputs} do not match"
                f" actual outputs {self.outputs}"
            )
            # We don't need to check the dims, dtype, etc, as this is already done in mark_as_output.

    def total_loss(self) -> Union[Tensor, float]:
        """
        :return: total loss, as it is used for backpropagation
        """
        assert self.stage == "train_step"
        assert self.losses, "call rf.get_run_ctx().mark_as_loss(...)"
        loss = 0.0
        for name, loss_obj in self.losses.items():
            if loss_obj.scale == 0.0 or loss_obj.as_error:
                continue
            loss += loss_obj.get_scaled_reduced_loss()
        return loss


@dataclass
class Loss:
    """
    Loss via :func:`RunCtx.mark_as_loss`.

    We collect all relevant information here.
    """

    loss: Tensor
    name: str

    scale: float = 1.0
    as_error: bool = False
    use_normalized_loss: bool = False  # for the gradient / total loss
    use_flatten_frames: bool = True
    custom_inv_norm_factor: Optional[Tensor] = None

    _summed_loss_cached: Optional[Tensor] = None
    _mean_loss_cached: Optional[Tensor] = None

    def get_summed_loss(self) -> Tensor:
        """
        :return: sum of loss (scalar)
        """
        if not self.loss.dims:
            return self.loss
        if self._summed_loss_cached is not None:
            return self._summed_loss_cached
        if self._mean_loss_cached is not None:
            return self._mean_loss_cached / self.get_inv_norm_factor()
        self._summed_loss_cached = rf.reduce_sum(self.loss, axis=self.loss.dims)
        return self._summed_loss_cached

    def get_mean_loss(self) -> Tensor:
        """
        :return: sum of loss (scalar)
        """
        if self._mean_loss_cached is not None:
            return self._mean_loss_cached
        if self.custom_inv_norm_factor:
            loss = self.get_summed_loss()
            loss /= rf.cast(self.custom_inv_norm_factor, dtype=loss.dtype)
            return loss
        if not self.loss.dims:
            return self.loss
        self._mean_loss_cached = rf.reduce_mean(self.loss, axis=self.loss.dims)
        return self._mean_loss_cached

    def get_inv_norm_factor(self) -> Union[int, Tensor]:
        """
        :return: inverse norm factor (scalar)
        """
        if self.custom_inv_norm_factor:
            return self.custom_inv_norm_factor
        return self.loss.num_elements()

    def get_scaled_reduced_loss(self) -> Tensor:
        """
        :return: scaled reduced loss (scalar), as it is supposed to be used for calculating the train gradient
        """
        if self.use_normalized_loss:
            loss = self.get_mean_loss()
        else:
            loss = self.get_summed_loss()
        return loss * self.scale


def _default_dim_order(tensor: Tensor) -> Sequence[Dim]:
    """
    See if some reasonable default dim order like BTF or BF is possible.

    :param tensor:
    :return:
    """
    rem_dims = list(tensor.dims)
    dims = []
    if tensor.have_batch_axis():
        rem_dims.remove(tensor.get_batch_dim_tag())
        dims.append(tensor.get_batch_dim_tag())
    if tensor.have_time_axis():
        rem_dims.remove(tensor.get_time_dim_tag())
        dims.append(tensor.get_time_dim_tag())
    dyn_dims = [d for d in rem_dims if d.is_dynamic()]
    if len(dyn_dims) > 1:
        raise Exception(
            f"Cannot infer order of dims automatically for output {tensor}. Please specify `dims` explicitly."
        )
    elif len(dyn_dims) == 1:
        rem_dims.remove(dyn_dims[0])
        dims.append(dyn_dims[0])
    if len(rem_dims) > 1:
        raise Exception(
            f"Cannot infer order of dims automatically for output {tensor}. Please specify `dims` explicitly."
        )
    elif len(rem_dims) == 1:
        dims.append(rem_dims[0])
    return dims



"""
Defines multiple signal processing-related layers.
"""

from __future__ import annotations

import tensorflow as tf
import returnn.tf.compat as tf_compat
from returnn.tf.layers.basic import (
    LayerBase,
    _ConcatInputLayer,
    ConvLayer,
    TransposedConvLayer,
    get_concat_sources_data_template,
)
from returnn.tf.util.basic import Data


class AlternatingRealToComplexLayer(_ConcatInputLayer):
    """
    This layer converts a real valued input tensor into a complex valued output tensor.
    For this even and odd features are considered the real and imaginary part of one complex number, respectively
    """

    layer_class = "alternating_real_to_complex"

    def __init__(self, **kwargs):
        super(AlternatingRealToComplexLayer, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        real_value = tf.strided_slice(input_placeholder, [0, 0, 0], tf.shape(input_placeholder), [1, 1, 2])
        imag_value = tf.strided_slice(input_placeholder, [0, 0, 1], tf.shape(input_placeholder), [1, 1, 2])
        self.output.placeholder = tf.complex(real_value, imag_value)
        self.output.size_placeholder = {
            0: self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch]
        }

    @classmethod
    def get_out_data_from_opts(cls, name, sources, n_out=None, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        return super(AlternatingRealToComplexLayer, cls).get_out_data_from_opts(
            name=name,
            sources=sources,
            out_type={"dim": n_out, "dtype": "complex64", "batch_dim_axis": 0, "time_dim_axis": 1},
            **kwargs,
        )


class BatchMedianPoolingLayer(_ConcatInputLayer):
    """
    This layer is used to pool together batches by taking their medium value.
    Thus the batch size is divided by pool_size.
    The stride is hard coded to be equal to the pool size.
    """

    layer_class = "batch_median_pooling"

    def __init__(self, pool_size=1, **kwargs):
        """
        :param int pool_size: size of the pool to take median of (is also used as stride size)
        """
        super(BatchMedianPoolingLayer, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        # get median over pooled batches
        # - reshape input for usage with tf.nn.top_k
        reshaped_input = tf.reshape(
            tf.transpose(input_placeholder, [1, 2, 0]),
            shape=(
                tf.shape(input_placeholder)[1],
                tf.shape(input_placeholder)[2],
                tf.shape(input_placeholder)[0] // pool_size,
                pool_size,
            ),
        )
        # - get median of each pool
        median = tf.nn.top_k(
            reshaped_input, k=tf.cast(tf_compat.v1.ceil(tf.constant(pool_size, dtype=tf.float32) / 2.0), dtype=tf.int32)
        ).values[:, :, :, -1]
        median_batch_major = tf.transpose(median, [2, 0, 1])
        self.output.placeholder = median_batch_major
        self.output.size_placeholder = {
            self.output.time_dim_axis_excluding_batch: tf.strided_slice(
                self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch],
                [0],
                tf.shape(self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch]),
                [pool_size],
            )
        }

    @classmethod
    def get_out_data_from_opts(cls, name, sources, pool_size, n_out=None, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param int pool_size:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        input_data = get_concat_sources_data_template(sources)
        assert not input_data.sparse
        return Data(
            name="%s_output" % name,
            shape=[
                input_data.get_placeholder_as_batch_major().shape[1].value,
                input_data.get_placeholder_as_batch_major().shape[2].value,
            ],
            dtype=input_data.dtype,
            size_placeholder={
                0: tf.strided_slice(
                    input_data.size_placeholder[input_data.time_dim_axis_excluding_batch],
                    [0],
                    tf.shape(input_data.size_placeholder[input_data.time_dim_axis_excluding_batch]),
                    [pool_size],
                )
            },
            sparse=False,
            batch_dim_axis=0,
            time_dim_axis=1,
        )


class ComplexLinearProjectionLayer(_ConcatInputLayer):
    """
    Complex linear projection layer
    For the original idea, see:
    Variani, Ehsan, et al. "Complex linear projection (CLP): A discriminative approach to joint feature extraction and
    acoustic modeling." (2016).
    """

    layer_class = "complex_linear_projection"

    def __init__(self, nr_of_filters, clp_weights_init="glorot_uniform", **kwargs):
        """
        :param int nr_of_filters:
        :param str|dict[str]|float|numpy.ndarray clp_weights_init:
        """
        if "n_out" in kwargs and kwargs["n_out"] != nr_of_filters:
            raise Exception("argument n_out of layer MelFilterbankLayer can not be different from nr_of_filters")
        kwargs["n_out"] = nr_of_filters
        self._nr_of_filters = nr_of_filters
        super(ComplexLinearProjectionLayer, self).__init__(**kwargs)
        self._clp_kernel = self._build_kernel(clp_weights_init)
        self.output.placeholder = self._build_clp_multiplication(self._clp_kernel)

    def _build_kernel(self, clp_weights_init):
        from returnn.tf.util.basic import get_initializer

        input_placeholder = self.input_data.get_placeholder_as_batch_major()
        kernel_width = input_placeholder.shape[2].value // 2
        kernel_height = self._nr_of_filters
        with self.var_creation_scope():
            clp_weights_initializer = get_initializer(
                clp_weights_init, seed=self.network.random.randint(2**31), eval_local_ns={"layer": self}
            )
            clp_kernel = self.add_param(
                tf_compat.v1.get_variable(
                    name="clp_kernel",
                    shape=(2, kernel_width, kernel_height),
                    dtype=tf.float32,
                    initializer=clp_weights_initializer,
                )
            )
        return clp_kernel

    def _build_clp_multiplication(self, clp_kernel):
        from returnn.tf.util.basic import safe_log

        input_placeholder = self.input_data.get_placeholder_as_batch_major()
        tf_compat.v1.assert_equal(tf.shape(clp_kernel)[1], tf.shape(input_placeholder)[2] // 2)
        tf_compat.v1.assert_equal(tf.shape(clp_kernel)[2], self._nr_of_filters)
        input_real = tf.strided_slice(input_placeholder, [0, 0, 0], tf.shape(input_placeholder), [1, 1, 2])
        input_imag = tf.strided_slice(input_placeholder, [0, 0, 1], tf.shape(input_placeholder), [1, 1, 2])
        kernel_real = self._clp_kernel[0, :, :]
        kernel_imag = self._clp_kernel[1, :, :]
        output_real = tf.einsum("btf,fp->btp", input_real, kernel_real) - tf.einsum(
            "btf,fp->btp", input_imag, kernel_imag
        )
        output_imag = tf.einsum("btf,fp->btp", input_imag, kernel_real) + tf.einsum(
            "btf,fp->btp", input_real, kernel_imag
        )
        output_uncompressed = tf.sqrt(tf.pow(output_real, 2) + tf.pow(output_imag, 2))
        output_compressed = safe_log(output_uncompressed)
        return output_compressed

    @classmethod
    def get_out_data_from_opts(cls, nr_of_filters, **kwargs):
        """
        :param int nr_of_filters:
        :rtype: Data
        """
        if "n_out" not in kwargs:
            kwargs["n_out"] = nr_of_filters
        return super(ComplexLinearProjectionLayer, cls).get_out_data_from_opts(**kwargs)


class ComplexToAlternatingRealLayer(_ConcatInputLayer):
    """
    This layer converts a complex valued input tensor into a real valued output tensor.
    For this the even and odd parts of the output are considered the real and imaginary part of one complex number,
    respectively.
    """

    layer_class = "complex_to_alternating_real"

    def __init__(self, **kwargs):
        def _interleave_vectors(vec1, vec2):
            vec1 = tf.expand_dims(vec1, 3)
            vec2 = tf.expand_dims(vec2, 3)
            interleaved = tf.concat([vec1, vec2], 3)
            interleaved = tf.reshape(interleaved, (tf.shape(vec1)[0], tf.shape(vec1)[1], tf.shape(vec1)[2] * 2))
            return interleaved

        super(ComplexToAlternatingRealLayer, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        real_value = tf_compat.v1.real(input_placeholder)
        imag_value = tf_compat.v1.imag(input_placeholder)
        self.output.placeholder = _interleave_vectors(real_value, imag_value)
        self.output.size_placeholder = {
            0: self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch]
        }


class MaskBasedGevBeamformingLayer(LayerBase):
    """
    This layer applies GEV beamforming to a multichannel signal.
    The different channels are assumed to be concatenated to
    the input feature vector. The first source to the layer must contain the complex spectrograms of the single channels
    and the second source must contain the noise and speech masks
    """

    layer_class = "mask_based_gevbeamforming"

    def __init__(self, nr_of_channels=1, postfilter_id=0, qralgorithm_steps=None, output_nan_filter=False, **kwargs):
        """
        :param int nr_of_channels: number of input channels to beamforming (needed to split the feature vector)
        :param int postfilter_id: Id which is specifying which post filter to apply in gev beamforming.
                                  For more information see
                                  tfSi6Proc.audioProcessing.enhancement.beamforming.TfMaskBasedGevBeamformer
        :param int|None: nr of steps of the qr algorithm to compute eigen vector for beamforming
        :param bool output_nan_filter: if set to true nan values in the beamforming output are replaced by zero
        """
        super(MaskBasedGevBeamformingLayer, self).__init__(**kwargs)
        assert len(self.sources) == 2
        # noinspection PyUnresolvedReferences
        from tfSi6Proc.audioProcessing.enhancement.beamforming import TfMaskBasedGevBeamformer

        complex_spectrogram = self.sources[0].output.get_placeholder_as_batch_major()
        complex_spectrogram = tf.transpose(
            tf.reshape(
                complex_spectrogram,
                (
                    tf.shape(complex_spectrogram)[0],
                    tf.shape(complex_spectrogram)[1],
                    nr_of_channels,
                    tf.shape(complex_spectrogram)[2] // nr_of_channels,
                ),
            ),
            [0, 1, 3, 2],
        )
        masks = tf.transpose(
            self.sources[1].output.placeholder,
            [
                self.sources[1].output.batch_dim_axis,
                self.sources[1].output.time_dim_axis,
                self.sources[1].output.feature_dim_axis,
            ],
        )
        masks = tf.transpose(
            tf.reshape(
                masks, (tf.shape(masks)[0], tf.shape(masks)[1], nr_of_channels, tf.shape(masks)[2] // nr_of_channels)
            ),
            [0, 1, 3, 2],
        )
        noise_masks = masks[:, :, : (tf.shape(masks)[2] // 2), :]
        speech_masks = masks[:, :, (tf.shape(masks)[2] // 2) :, :]

        gev_bf = TfMaskBasedGevBeamformer(
            flag_inputHasBatch=1,
            tfFreqDomInput=complex_spectrogram,
            tfNoiseMask=noise_masks,
            tfSpeechMask=speech_masks,
            postFilterId=postfilter_id,
            qrAlgorithmSteps=qralgorithm_steps,
            outputNanFilter=output_nan_filter,
        )
        bf_out = gev_bf.getFrequencyDomainOutputSignal()
        self.output.placeholder = bf_out

    @classmethod
    def get_out_data_from_opts(cls, out_type=None, n_out=None, **kwargs):
        """
        :param dict[str]|None out_type:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        if out_type is None:
            out_type = {}
        out_type.setdefault("dim", n_out)
        out_type["batch_dim_axis"] = 0
        out_type["time_dim_axis"] = 1
        return super(MaskBasedGevBeamformingLayer, cls).get_out_data_from_opts(out_type=out_type, **kwargs)


class MaskBasedMvdrBeamformingWithDiagLoadingLayer(LayerBase):
    """
    This layer applies GEV beamforming to a multichannel signal.
    The different channels are assumed to be concatenated to
    the input feature vector.
    The first source to the layer must contain the complex spectrograms of the single channels
    and the second source must contain the noise and speech masks.
    """

    layer_class = "mask_based_mvdrbeamforming"

    def __init__(
        self, nr_of_channels=1, diag_loading_coeff=0, qralgorithm_steps=None, output_nan_filter=False, **kwargs
    ):
        """
        :param int nr_of_channels: number of input channels to beamforming (needed to split the feature vector)
        :param int diag_loading_coeff: weighting coefficient for diagonal loading.
        :param int|None qralgorithm_steps: nr of steps of the qr algorithm to compute eigen vector for beamforming
        :param bool output_nan_filter: if set to true nan values in the beamforming output are replaced by zero
        """
        super(MaskBasedMvdrBeamformingWithDiagLoadingLayer, self).__init__(**kwargs)
        assert len(self.sources) == 2

        # noinspection PyUnresolvedReferences
        from tfSi6Proc.audioProcessing.enhancement.beamforming import TfMaskBasedMvdrBeamformer

        complex_spectrogram_with_concat_channels = self.sources[0].output.get_placeholder_as_batch_major()
        complex_spectrogram = tf.transpose(
            tf.reshape(
                complex_spectrogram_with_concat_channels,
                (
                    tf.shape(complex_spectrogram_with_concat_channels)[0],
                    tf.shape(complex_spectrogram_with_concat_channels)[1],
                    nr_of_channels,
                    tf.shape(complex_spectrogram_with_concat_channels)[2] // nr_of_channels,
                ),
            ),
            [0, 1, 3, 2],
        )
        noise_masks = self.sources[1].output.get_placeholder_as_batch_major()
        noise_masks = tf.transpose(
            tf.reshape(
                noise_masks,
                (
                    tf.shape(noise_masks)[0],
                    tf.shape(noise_masks)[1],
                    nr_of_channels,
                    tf.shape(noise_masks)[2] // nr_of_channels,
                ),
            ),
            [0, 1, 3, 2],
        )

        mvdr_bf = TfMaskBasedMvdrBeamformer(
            flag_inputHasBatch=1,
            tfFreqDomInput=complex_spectrogram,
            tfNoiseMask=noise_masks,
            tfDiagLoadingCoeff=tf.constant(diag_loading_coeff, dtype=tf.float32),
            qrAlgorithmSteps=qralgorithm_steps,
            outputNanFilter=output_nan_filter,
        )
        bf_out = mvdr_bf.getFrequencyDomainOutputSignal()
        self.output.placeholder = bf_out

    @classmethod
    def get_out_data_from_opts(cls, out_type=None, n_out=None, **kwargs):
        """
        :param dict[str]|None out_type:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        if out_type is None:
            out_type = {}
        out_type.setdefault("dim", n_out)
        out_type["batch_dim_axis"] = 0
        out_type["time_dim_axis"] = 1
        return super(MaskBasedMvdrBeamformingWithDiagLoadingLayer, cls).get_out_data_from_opts(
            out_type=out_type, **kwargs
        )


class MelFilterbankLayer(_ConcatInputLayer):
    """
    This layer applies the Mel filterbank to the input.
    """

    layer_class = "mel_filterbank"

    def __init__(self, sampling_rate=16000, fft_size=1024, nr_of_filters=80, f_min=None, f_max=None, **kwargs):
        """
        :param int sampling_rate: sampling rate of the signal which the input originates from
        :param int fft_size: fft_size with which the time signal was transformed into the intput
        :param int nr_of_filters: number of output filter bins
        :param float f_min: minimum frequency for mel filters
        :param float f_max: maximum frequency for mel filters
        """

        # noinspection PyShadowingNames
        def tf_mel_filter_bank(f_min, f_max, sampling_rate, fft_size, nr_of_filters):
            """
            Returns the filter matrix which yields the mel filter bank features, when applied to the spectrum as
            tf.matmul(freqDom, filterMatrix), where freqDom has dimension (time, frequency)
            and filterMatrix is the matrix returned
            by this function
            The filter matrix is computed according to equation 6.141 in
            [Huang & Acero+, 2001] "Spoken Language Processing - A Guide to Theroy, Algorithm, and System Development"

            :param float|int f_min: minimum frequency
            :param float|int f_max: maximum frequency
            :param float sampling_rate: sampling rate of audio signal
            :param int fft_size: dimension of discrete fourier transformation
            :param int nr_of_filters: number of mel frequency filter banks to be created
            :rtype: tf.tensor, shape=(filterValue, nrOfFilters)
            :return: matrix yielding the mel frequency cepstral coefficients
            """
            import numpy as np

            def mel_scale(freq):
                """
                returns the respective value on the mel scale

                :param float freq: frequency value to transform onto mel scale
                :rtype: float
                """
                return 1125.0 * np.log(1 + float(freq) / 700)

            def inv_mel_scale(mel_val):
                """
                returns the respective value in the frequency domain

                :param float mel_val: value in mel domain
                :rtype: float
                """
                return 700.0 * (np.exp(float(mel_val) / 1125) - 1)

            # noinspection PyShadowingNames
            def filter_center(filter_id, f_min, f_max, sampling_rate, fft_size, nr_of_filters):
                """
                :param int filter_id: filter to compute the center frequency for
                :param float|int f_min: minimum frequency
                :param float|int f_max: maximum frequency
                :param float sampling_rate: sampling rate of audio signal
                :param int fft_size: dimension of discrete fourier transformation
                :param int nr_of_filters: number of mel frequency filter banks to be created
                :rtype: float
                :return: center frequency of filter
                """
                return (float(fft_size) / sampling_rate) * inv_mel_scale(
                    mel_scale(f_min) + filter_id * ((mel_scale(f_max) - mel_scale(f_min)) / (nr_of_filters + 1))
                )

            filt_cent = np.zeros(shape=(nr_of_filters + 2,), dtype=np.float32)
            for i1 in range(nr_of_filters + 2):
                filt_cent[i1] = filter_center(i1, f_min, f_max, sampling_rate, fft_size, nr_of_filters)
            f_mat = np.zeros(shape=(int(np.floor(fft_size / 2) + 1), nr_of_filters))
            for i1 in range(f_mat.shape[0]):
                for i2 in range(1, nr_of_filters + 1):
                    if (i1 > filt_cent[i2 - 1]) and (i1 < filt_cent[i2 + 1]):
                        if i1 < filt_cent[i2]:
                            num = i1 - filt_cent[i2 - 1]
                            denom = filt_cent[i2] - filt_cent[i2 - 1]
                        else:
                            num = filt_cent[i2 + 1] - i1
                            denom = filt_cent[i2 + 1] - filt_cent[i2]
                        el_val = num / denom
                    else:
                        el_val = 0
                    f_mat[i1, i2 - 1] = el_val
            return tf.constant(f_mat, dtype=tf.float32)

        if "n_out" in kwargs and (kwargs["n_out"] != nr_of_filters):
            raise Exception("argument n_out of layer MelFilterbankLayer can not be different from nr_of_filters")
        kwargs["n_out"] = nr_of_filters
        super(MelFilterbankLayer, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        f_min = f_min or 0
        f_max = f_max or sampling_rate / 2.0
        mel_fbank_mat = tf_mel_filter_bank(f_min, f_max, sampling_rate, fft_size, nr_of_filters)
        self.output.placeholder = tf.einsum(
            "btf,bfc->btc",
            input_placeholder,
            tf.tile(tf.expand_dims(mel_fbank_mat, axis=0), [tf.shape(input_placeholder)[0], 1, 1]),
        )
        self.output.size_placeholder = {
            0: self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch]
        }

    @classmethod
    def get_out_data_from_opts(cls, name, sources, n_out=None, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        return super(MelFilterbankLayer, cls).get_out_data_from_opts(
            name=name, sources=sources, out_type={"dim": n_out, "batch_dim_axis": 0, "time_dim_axis": 1}, **kwargs
        )


class MultiChannelMultiResolutionStftLayer(_ConcatInputLayer):
    """
    The layer applys a STFT to every channel separately and concatenates the frequency domain vectors for every frame.
    The STFT is applied with multiple different frame- and FFT-sizes and the resulting multi-channel STFTs are
    concatenated.
    Resulting in a tensor with the content [res_0-ch_0, ..., res_0-ch_N, res_1-ch_0, ... res_M-ch_N]
    The subsampling from T input samples to T' output frames is computed as follows:
    T' = (T - frame_size) / frame_shift + 1
    frame_shift is the same for all resolutions and T' is computed according to a reference frame_size which is taken to
    be frame_sizes[0].
    For all other frame sizes the input is zero-padded or the output is cut to obtain the same T' as for the reference
    frame_size.
    """

    layer_class = "multichannel_multiresolution_stft_layer"
    recurrent = True

    def __init__(
        self,
        frame_shift,
        frame_sizes,
        fft_sizes,
        window="hanning",
        use_rfft=True,
        nr_of_channels=1,
        pad_last_frame=False,
        **kwargs,
    ):
        """
        :param int frame_shift: frame shift for stft in samples
        :param list[int] frame_sizes: frame size for stft in samples
        :param list[int] fft_sizes: fft size in samples
        :param str window: id of the windowing function used. Possible options are:
            - hanning
        :param bool use_rfft: if set to true a real input signal is expected
            and only the significant half of the FFT bins are returned
        :param int nr_of_channels: number of input channels
        :param bool pad_last_frame: padding of last frame with zeros or discarding of last frame
        """

        def _compute_size_placeholder():
            size_placeholder_dict = {}
            if self._pad_last_frame:
                new_size = (
                    self.input_data.size_placeholder[0] - self._reference_frame_size - 1
                ) // self._frame_shift + 2
            else:
                new_size = (self.input_data.size_placeholder[0] - self._reference_frame_size) // self._frame_shift + 1
            from ..util.data import Dim

            Dim(
                kind=Dim.Types.Spatial,
                description="%s:MultiChannelMultiResolutionStft" % self.name,
                dyn_size=new_size,
                batch=self.output.batch,
                src_data=self.output,
                src_axis=self.output.get_batch_axis(0),
                auto_generated=True,
                dimension=None,
            )
            size_placeholder_dict[0] = new_size
            return size_placeholder_dict

        n_out = sum([self._get_n_out_by_fft_config(fft_size, use_rfft, nr_of_channels) for fft_size in fft_sizes])
        if "n_out" in kwargs and (kwargs["n_out"] != n_out):
            raise Exception("argument n_out of layer MultiChannelStftLayer does not match the fft configuration")
        kwargs["n_out"] = n_out
        super(MultiChannelMultiResolutionStftLayer, self).__init__(**kwargs)
        tf_compat.v1.assert_equal(nr_of_channels, self._get_nr_of_channels_from_input_placeholder())
        self._nr_of_channels = nr_of_channels
        self._frame_shift = frame_shift
        self._frame_sizes = frame_sizes
        self._reference_frame_size = frame_sizes[0]
        self._fft_sizes = fft_sizes
        self._window = window
        self._use_rfft = use_rfft
        self._pad_last_frame = pad_last_frame
        self.output.placeholder = self._apply_stft_to_input()
        self.output.size_placeholder = _compute_size_placeholder()

    def _get_nr_of_channels_from_input_placeholder(self):
        input_placeholder = self.input_data.get_placeholder_as_batch_major()
        return input_placeholder.shape[2]

    def _apply_stft_to_input(self):
        from returnn.tf.util.basic import get_shape

        # noinspection PyShadowingNames
        def _crop_stft_output_to_reference_frame_size_length(channel_concatenated_stft, crop_size):
            return tf.slice(
                channel_concatenated_stft,
                [0, 0, 0],
                [get_shape(channel_concatenated_stft)[0], crop_size, get_shape(channel_concatenated_stft)[2]],
            )

        input_placeholder = self.input_data.get_placeholder_as_batch_major()
        channel_wise_stft_res_list = list()
        for fft_size, frame_size in zip(self._fft_sizes, self._frame_sizes):

            def _get_window(window_length, dtype):
                if self._window == "hanning":
                    try:
                        # noinspection PyPackageRequirements
                        from tensorflow.signal import hann_window
                    except ImportError:
                        # noinspection PyPackageRequirements,PyUnresolvedReferences
                        from tensorflow.contrib.signal import hann_window
                    window = hann_window(window_length, dtype=dtype)
                elif self._window == "blackman":
                    # noinspection PyPackageRequirements
                    import scipy.signal

                    window = tf.constant(scipy.signal.windows.blackman(frame_size), dtype=tf.float32)
                elif self._window == "None" or self._window == "ones":
                    window = tf.ones((window_length,), dtype=dtype)
                else:
                    assert False, "Window was not parsed correctly: {}".format(self._window)
                return window

            # noinspection PyShadowingNames
            def _pad_time_signal(input_placeholder, frame_size):
                if frame_size > self._reference_frame_size:
                    return tf.concat(
                        [
                            input_signal,
                            tf.ones(
                                [
                                    get_shape(input_signal)[0],
                                    frame_size - self._reference_frame_size,
                                    get_shape(input_signal)[2],
                                ]
                            )
                            * 1e-7,
                        ],
                        axis=1,
                    )
                else:
                    return input_placeholder

            input_signal = _pad_time_signal(input_placeholder, frame_size)
            if self._use_rfft:
                try:
                    # noinspection PyPackageRequirements
                    from tensorflow.signal import stft
                except ImportError:
                    # noinspection PyPackageRequirements,PyUnresolvedReferences
                    from tensorflow.contrib.signal import stft
                channel_wise_stft = stft(
                    signals=tf.transpose(input_signal, [0, 2, 1]),
                    frame_length=frame_size,
                    frame_step=self._frame_shift,
                    fft_length=fft_size,
                    window_fn=_get_window,
                    pad_end=self._pad_last_frame,
                )
                channel_wise_stft = tf.transpose(channel_wise_stft, [0, 2, 1, 3])
                batch_dim = tf.shape(channel_wise_stft)[0]
                time_dim = tf.shape(channel_wise_stft)[1]
                concat_feature_dim = channel_wise_stft.shape[2] * channel_wise_stft.shape[3]
                channel_concatenated_stft = tf.reshape(channel_wise_stft, (batch_dim, time_dim, concat_feature_dim))
                if channel_wise_stft_res_list:
                    channel_concatenated_stft = _crop_stft_output_to_reference_frame_size_length(
                        channel_concatenated_stft, get_shape(channel_wise_stft_res_list[0])[1]
                    )
                channel_wise_stft_res_list.append(channel_concatenated_stft)
        output_placeholder = tf.concat(channel_wise_stft_res_list, axis=2)
        return output_placeholder

    @classmethod
    def _get_n_out_by_fft_config(cls, fft_size, use_rfft, nr_of_channels):
        n_out = fft_size
        if use_rfft:
            n_out = fft_size // 2 + 1
        n_out *= nr_of_channels
        return n_out

    @classmethod
    def get_out_data_from_opts(cls, fft_sizes, use_rfft=True, nr_of_channels=1, **kwargs):
        """
        :param list[int] fft_sizes:
        :param bool use_rfft:
        :param int nr_of_channels:
        :rtype: Data
        """
        n_out = sum([cls._get_n_out_by_fft_config(fft_size, use_rfft, nr_of_channels) for fft_size in fft_sizes])
        if "n_out" not in kwargs:
            kwargs["n_out"] = n_out
        return (
            super(MultiChannelMultiResolutionStftLayer, cls)
            .get_out_data_from_opts(**kwargs)
            .copy_template(dtype="complex64")
        )


class MultiChannelStftLayer(MultiChannelMultiResolutionStftLayer):
    """
    The layer applys a STFT to every channel separately and concatenates the frequency domain vectors for every frame.
    """

    recurrent = True
    layer_class = "multichannel_stft_layer"

    def __init__(
        self,
        frame_shift,
        frame_size,
        fft_size,
        window="hanning",
        use_rfft=True,
        nr_of_channels=1,
        pad_last_frame=False,
        **kwargs,
    ):
        kwargs["frame_shift"] = frame_shift
        kwargs["window"] = window
        kwargs["use_rfft"] = use_rfft
        kwargs["nr_of_channels"] = nr_of_channels
        kwargs["pad_last_frame"] = pad_last_frame
        super(MultiChannelStftLayer, self).__init__(frame_sizes=[frame_size], fft_sizes=[fft_size], **kwargs)

    @classmethod
    def get_out_data_from_opts(cls, fft_size, use_rfft=True, nr_of_channels=1, **kwargs):
        """
        :param int fft_size:
        :param bool use_rfft:
        :param int nr_of_channels:
        :rtype: Data
        """
        return (
            super(MultiChannelStftLayer, cls)
            .get_out_data_from_opts(fft_sizes=[fft_size], use_rfft=use_rfft, nr_of_channels=nr_of_channels, **kwargs)
            .copy_template(dtype="complex64")
        )


class StftLayer(_ConcatInputLayer):
    """
    A generic STFT layer.
    """

    layer_class = "stft"
    recurrent = True  # we should not shuffle in the time-dimension

    # noinspection PyUnusedLocal
    def __init__(
        self,
        frame_shift,
        frame_size,
        fft_size=None,
        in_spatial_dims=None,
        out_spatial_dims=None,
        out_dim=None,
        **kwargs,
    ):
        """
        :param int frame_shift: frame shift for STFT
        :param int frame_size: frame size for STFT
        :param Optional[int] fft_size: size of the FFT to apply
        :param list[Dim|str]|None in_spatial_dims:
        :param list[Dim]|None out_spatial_dims:
        :param Dim|None out_dim:
        """
        fft_size = fft_size or frame_size
        assert "n_out" not in kwargs
        assert "out_type" not in kwargs
        super(StftLayer, self).__init__(**kwargs)
        assert not self.input_data.sparse
        if in_spatial_dims:
            assert len(in_spatial_dims) == 1
            axis = self.input_data.get_axis_from_description(in_spatial_dims[0])
            assert axis is not None
            self.input_data.version = 1
            self.input_data.time_dim_axis = axis
        assert self.input_data.have_batch_axis()
        assert self.input_data.have_time_axis()
        in_dim = self.input_data.feature_dim_or_sparse_dim
        input_data, num_batch_dims = ConvLayer.transform_input(
            self.input_data, network=self.network, in_dim=in_dim, in_spatial_dims=in_spatial_dims
        )
        x = input_data.placeholder
        if input_data.have_feature_axis():
            # squeeze feature axis
            assert input_data.batch_shape[input_data.feature_dim_axis] == 1, "only implemented for single channel"
            x = tf.squeeze(x, axis=input_data.feature_dim_axis)
        extended_batch_shape = None
        if num_batch_dims > 1:
            x_shape = tf.shape(x)
            extended_batch_shape = x_shape[:num_batch_dims]
            x = tf.reshape(x, tf.concat([[-1], x_shape[num_batch_dims:]], axis=0))  # merge all batch dims
        y = tf.signal.stft(x, frame_length=frame_size, frame_step=frame_shift, fft_length=fft_size)
        # y shape is [batch] + spatial_dims + [n_out].
        if num_batch_dims > 1:
            y = tf.reshape(y, tf.concat([extended_batch_shape, tf.shape(y)[1:]], axis=0))
        self.output.placeholder = y

    @classmethod
    def get_out_data_from_opts(
        cls,
        name,
        sources,
        network,
        frame_shift,
        frame_size,
        fft_size=None,
        in_spatial_dims=None,
        out_spatial_dims=None,
        out_dim=None,
        **kwargs,
    ):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param returnn.tf.network.TFNetwork network:
        :param int frame_shift: frame shift for STFT
        :param int frame_size: frame size for STFT
        :param Optional[int] fft_size: size of the FFT to apply
        :param list[Dim|str]|None in_spatial_dims:
        :param list[Dim]|None out_spatial_dims:
        :param Dim|None out_dim:
        :rtype: Data
        """
        fft_size = fft_size or frame_size
        data = ConvLayer.get_out_data_from_opts(
            name=name,
            sources=sources,
            network=network,
            filter_size=[frame_size],
            strides=[frame_shift],
            padding="valid",
            n_out=fft_size // 2 + 1,
            in_spatial_dims=in_spatial_dims,
            out_spatial_dims=out_spatial_dims,
            out_dim=out_dim,
        )
        data = data.copy_with_feature_dim_axis(-1)

        data.dtype = "complex64"
        return data


class IstftLayer(_ConcatInputLayer):
    """
    A generic iSTFT layer.
    """

    layer_class = "istft"
    recurrent = True  # we should not shuffle in the time-dimension

    # noinspection PyUnusedLocal
    def __init__(
        self,
        frame_shift,
        frame_size,
        fft_size=None,
        in_spatial_dims=None,
        out_spatial_dims=None,
        out_dim=None,
        **kwargs,
    ):
        """
        :param int frame_shift: frame shift for STFT
        :param int frame_size: frame size for STFT
        :param Optional[int] fft_size: size of the FFT to apply
        :param list[Dim|str]|None in_spatial_dims:
        :param list[Dim]|None out_spatial_dims:
        :param Dim|None out_dim:
        """
        fft_size = fft_size or frame_size
        assert "n_out" not in kwargs
        assert "out_type" not in kwargs
        super(IstftLayer, self).__init__(**kwargs)
        assert not self.input_data.sparse
        assert self.input_data.have_batch_axis()
        assert self.input_data.have_time_axis()
        assert self.input_data.have_feature_axis()
        in_dim = self.input_data.feature_dim_or_sparse_dim
        assert in_dim.dimension == fft_size // 2 + 1
        input_data, num_batch_dims = ConvLayer.transform_input(
            self.input_data, network=self.network, in_dim=in_dim, in_spatial_dims=in_spatial_dims
        )
        # shape should be (B, ..., T, F)
        input_data = input_data.copy_with_feature_dim_axis(-1)
        input_data = input_data.copy_with_time_dim_axis(-2)
        x = input_data.placeholder
        extended_batch_shape = None
        if num_batch_dims > 1:
            x_shape = tf.shape(x)
            extended_batch_shape = x_shape[:num_batch_dims]
            x = tf.reshape(x, tf.concat([[-1], x_shape[num_batch_dims:]], axis=0))  # merge all batch dims
        window_fn = tf.signal.inverse_stft_window_fn(frame_shift)
        y = tf.signal.inverse_stft(
            x, frame_length=frame_size, frame_step=frame_shift, fft_length=fft_size, window_fn=window_fn
        )
        y = tf.expand_dims(y, -1)
        # y shape is [batch] + spatial_dims + [n_out].
        if num_batch_dims > 1:
            y = tf.reshape(y, tf.concat([extended_batch_shape, tf.shape(y)[1:]], axis=0))
        self.output.placeholder = y

    @classmethod
    def get_out_data_from_opts(
        cls,
        name,
        sources,
        network,
        frame_shift,
        frame_size,
        in_spatial_dims=None,
        out_spatial_dims=None,
        out_dim=None,
        **kwargs,
    ):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param returnn.tf.network.TFNetwork network:
        :param int frame_shift: frame shift for STFT
        :param int frame_size: frame size for STFT
        :param list[Dim|str]|None in_spatial_dims:
        :param list[Dim]|None out_spatial_dims:
        :param Dim|None out_dim:
        :rtype: Data
        """
        data = TransposedConvLayer.get_out_data_from_opts(
            name=name,
            sources=sources,
            network=network,
            filter_size=[frame_size],
            strides=[frame_shift],
            padding="valid",
            n_out=1,
            in_spatial_dims=in_spatial_dims,
            out_spatial_dims=out_spatial_dims,
            out_dim=out_dim,
        )
        data = data.copy_with_feature_dim_axis(-1)
        return data


class NoiseEstimationByFirstTFramesLayer(_ConcatInputLayer):
    """
    Estimates noise from the first t time frames.
    """

    layer_class = "first_t_frames_noise_estimator"
    recurrent = True

    def __init__(self, nr_of_frames, **kwargs):
        """
        :param int nr_of_frames: first nr_of_frames frames are used for averaging
                                 all frames are used if nr_of_frames is -1
        """
        super(NoiseEstimationByFirstTFramesLayer, self).__init__(**kwargs)
        self._nr_of_frames = nr_of_frames
        noise_vector = self._get_noise_vector()
        self.output.placeholder = tf.tile(
            noise_vector, (1, tf.shape(self.input_data.get_placeholder_as_batch_major())[1], 1)
        )

    def _get_noise_vector(self):
        input_placeholder = self.input_data.get_placeholder_as_batch_major()
        if self._nr_of_frames != -1:
            noise_vector = tf.reduce_mean(input_placeholder[:, : self._nr_of_frames, :], axis=1, keepdims=True)
        else:
            noise_vector = tf.reduce_mean(input_placeholder, axis=1, keepdims=True)
        return noise_vector


class ParametricWienerFilterLayer(LayerBase):
    """
    Parametric Wiener Filter
    For related paper, see:
    Menne, Tobias, Ralf Schlueter, and Hermann Ney.
    "Investigation into joint optimization of single channel speech enhancement and acoustic modeling for robust ASR."
    IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP).
    IEEE, 2019.
    """

    layer_class = "parametric_wiener_filter"

    def __init__(
        self,
        l_overwrite=None,
        p_overwrite=None,
        q_overwrite=None,
        filter_input=None,
        parameters=None,
        noise_estimation=None,
        average_parameters=False,
        **kwargs,
    ):
        """
        :param float|None l_overwrite:
            if given overwrites the l value of the parametric wiener filter with given constant
        :param float|None p_overwrite:
            if given overwrites the p value of the parametric wiener filter with given constant
        :param float|None q_overwrite:
            if given overwrites the q value of the parametric wiener filter with given constant
        :param LayerBase|None filter_input: name of layer containing input for wiener filter
        :param LayerBase|None parameters: name of layer containing parameters for wiener filter
        :param LayerBase|None noise_estimation: name of layer containing noise estimate for wiener filter
        :param bool average_parameters: if set to true the parameters l, p and q are averaged over the time axis
        """
        # noinspection PyUnresolvedReferences
        from tfSi6Proc.audioProcessing.enhancement.singleChannel import TfParametricWienerFilter

        super(ParametricWienerFilterLayer, self).__init__(**kwargs)

        class _NoiseEstimator(object):
            """
            Noise estimator. For details, see tfSi6Proc.audioProcessing.enhancement.singleChannel
            """

            def __init__(self, noise_power_spectrum_tensor):
                self._noise_power_spectrum_tensor = noise_power_spectrum_tensor

            @classmethod
            def from_layer(cls, layer):
                """
                :param LayerBase layer:
                :rtype: _NoiseEstimator
                """
                return cls(layer.output.get_placeholder_as_batch_major())

            def get_noise_power_spectrum(self):
                """
                :rtype: tf.Tensor
                """
                return self._noise_power_spectrum_tensor

        # noinspection PyShadowingNames
        def _get_parameters_from_constructor_inputs(
            parameters, l_overwrite, p_overwrite, q_overwrite, average_parameters
        ):
            parameter_vector = None
            if parameters is not None:
                parameter_vector = parameters.output.get_placeholder_as_batch_major()
                tf_compat.v1.assert_equal(parameter_vector.shape[-1], 3)
            if (l_overwrite is None) or (p_overwrite is None) or (q_overwrite is None):
                assert parameter_vector is not None
                if average_parameters:
                    parameter_vector = tf.tile(
                        tf.reduce_mean(parameter_vector, axis=1, keepdims=True), [1, tf.shape(parameter_vector)[1], 1]
                    )
            if l_overwrite is not None:
                l_ = tf.constant(l_overwrite, dtype=tf.float32)
            else:
                l_ = tf.expand_dims(parameter_vector[:, :, 0], axis=-1)
            if p_overwrite is not None:
                p_ = tf.constant(p_overwrite, dtype=tf.float32)
            else:
                p_ = tf.expand_dims(parameter_vector[:, :, 1], axis=-1)
            if q_overwrite is not None:
                q_ = tf.constant(q_overwrite, dtype=tf.float32)
            else:
                q_ = tf.expand_dims(parameter_vector[:, :, 2], axis=-1)
            return l_, p_, q_

        filter_input_placeholder = filter_input.output.get_placeholder_as_batch_major()
        if filter_input_placeholder.dtype != tf.complex64:
            filter_input_placeholder = tf.cast(filter_input_placeholder, dtype=tf.complex64)
        ne = _NoiseEstimator.from_layer(noise_estimation)
        l, p, q = _get_parameters_from_constructor_inputs(
            parameters, l_overwrite, p_overwrite, q_overwrite, average_parameters
        )
        wiener = TfParametricWienerFilter(ne, [], l, p, q, inputTensorFreqDomain=filter_input_placeholder)
        self.output.placeholder = wiener.getFrequencyDomainOutputSignal()

    @classmethod
    def get_out_data_from_opts(cls, **kwargs):
        """
        :rtype: Data
        """
        kwargs_with_sources = kwargs
        if ("sources" in kwargs_with_sources) and (len(kwargs_with_sources["sources"]) == 0):
            kwargs_with_sources["sources"] = [kwargs["filter_input"]]
        return cls._base_get_out_data_from_opts(**kwargs_with_sources)

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """
        :param dict[str] d:
        :param TFNetwork.TFNetwork network:
        :param ((str) -> LayerBase) get_layer:
        """
        if "from" in d and len(d["from"]) > 0:
            # This if block is kept for backwards compatibility only and should not be used
            assert ("filter_input" not in d) and ("parameters" not in d) and ("noise_estimation" not in d)
            if len(d["from"]) == 2:
                d["filter_input"] = d["from"][0]
                d["parameters"] = None
                d["noise_estimation"] = d["from"][1]
            if len(d["from"]) == 3:
                d["filter_input"] = d["from"][0]
                d["parameters"] = d["from"][1]
                d["noise_estimation"] = d["from"][2]
        d.setdefault("from", [])
        super(ParametricWienerFilterLayer, cls).transform_config_dict(d, network=network, get_layer=get_layer)
        d["filter_input"] = get_layer(d["filter_input"])
        if d["parameters"] is not None:
            d["parameters"] = get_layer(d["parameters"])
        d["noise_estimation"] = get_layer(d["noise_estimation"])


class SignalMaskingLayer(LayerBase):
    """
    Mask a given signal using a given mask.
    """

    layer_class = "signal_masking"

    def __init__(self, signal, mask, **kwargs):
        """
        :param LayerBase signal: name of layer the signal to be masked
        :param LayerBase mask: name of layer containing the mask
        """

        # noinspection PyShadowingNames
        def _cast_signal_and_mask_if_iecessary(signal, mask):
            if signal.dtype != mask.dtype:
                if signal.dtype == tf.complex64 and mask.dtype == tf.float32:
                    return signal, tf.cast(mask, dtype=tf.complex64)
                else:
                    raise NotImplementedError("difference in dtype between mask and signal is not supported yet.")
            return signal, mask

        super(SignalMaskingLayer, self).__init__(**kwargs)
        self._signal = signal.output.get_placeholder_as_batch_major()
        self._mask = mask.output.get_placeholder_as_batch_major()
        self._signal, self._mask = _cast_signal_and_mask_if_iecessary(self._signal, self._mask)
        self.output.placeholder = tf.multiply(self._signal, self._mask)
        self.output.size_placeholder = signal.output.size_placeholder

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """
        :param dict[str] d:
        :param TFNetwork.TFNetwork network:
        :param ((str) -> LayerBase) get_layer:
        """
        d.setdefault("from", [])
        super(SignalMaskingLayer, cls).transform_config_dict(d, network=network, get_layer=get_layer)
        d["signal"] = get_layer(d["signal"])
        d["mask"] = get_layer(d["mask"])


class SplitConcatMultiChannel(_ConcatInputLayer):
    """
    This layer assumes the feature vector to be a concatenation of features of multiple channels (of the same size).
    It splits the feature dimension into equisized number of channel features and stacks them in the batch dimension.
    Thus the batch size is multiplied with the number of channels and the feature size is divided by the number of
    channels.
    The channels of one singal will have consecutive batch indices, meaning the signal of the original batch index n is
    split and can now be found in batch indices (n * nr_of_channels) to ((n+1) * nr_of_channels - 1)
    """

    layer_class = "split_concatenated_multichannel"

    def __init__(self, nr_of_channels=1, **kwargs):
        """
        :param int nr_of_channels: the number of concatenated channels in the feature dimension
        """
        super(SplitConcatMultiChannel, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        # For shape comments: C = nr_of_channels
        output = tf.reshape(
            input_placeholder,
            [
                tf.shape(input_placeholder)[0],
                tf.shape(input_placeholder)[1],
                nr_of_channels,
                tf.shape(input_placeholder)[2] // nr_of_channels,
            ],
        )  # shape B x T x C x (F / C)
        new_shape = (tf.shape(output)[1], tf.shape(output)[3], tf.shape(output)[0] * tf.shape(output)[2])
        output = tf.transpose(output, [1, 3, 0, 2])  # shape T x (F / C) x B x C
        output = tf.reshape(output, new_shape)  # shape T x (F / C) x (B * C)
        self.output.placeholder = tf.transpose(output, [2, 0, 1])  # shape (B * C) x T x (F / C)
        # work around to obtain result like numpy.repeat(size_placeholder, nr_of_channels)
        size_placeholder_np_repeat = tf.reshape(
            self.input_data.size_placeholder[self.input_data.time_dim_axis_excluding_batch], [-1, 1]
        )  # shape T x 1
        size_placeholder_np_repeat = tf.tile(size_placeholder_np_repeat, [1, nr_of_channels])  # shape T x C
        size_placeholder_np_repeat = tf.reshape(size_placeholder_np_repeat, [-1])  # shape (T * C)
        self.output.size_placeholder = {self.output.time_dim_axis_excluding_batch: size_placeholder_np_repeat}

    @classmethod
    def get_out_data_from_opts(cls, name, sources, nr_of_channels, n_out=None, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param int nr_of_channels:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        input_data = get_concat_sources_data_template(sources).copy_as_batch_major()
        assert not input_data.sparse
        return Data(
            name="%s_output" % name,
            shape=[input_data.batch_shape[1], input_data.batch_shape[2] // nr_of_channels],
            dtype=input_data.dtype,
            sparse=False,
            batch_dim_axis=0,
            time_dim_axis=1,
        )


class TileFeaturesLayer(_ConcatInputLayer):
    """
    This function is tiling features with given number of repetitions.
    """

    layer_class = "tile_features"

    def __init__(self, repetitions=1, **kwargs):
        """
        :param int repetitions: number of tiling repetitions in feature domain
        """
        super(TileFeaturesLayer, self).__init__(**kwargs)

        input_placeholder = self.input_data.get_placeholder_as_batch_major()

        self.output.placeholder = tf.tile(input_placeholder, [1, 1, repetitions])

    @classmethod
    def get_out_data_from_opts(cls, name, sources, repetitions, n_out=None, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param int repetitions:
        :param int|None|returnn.util.basic.NotSpecified n_out:
        :rtype: Data
        """
        input_data = get_concat_sources_data_template(sources)
        assert not input_data.sparse
        return Data(
            name="%s_output" % name,
            shape=[input_data.shape[1], input_data.shape[2] * repetitions],
            dtype=input_data.dtype,
            sparse=False,
            size_placeholder={0: input_data.size_placeholder[input_data.time_dim_axis_excluding_batch]},
            batch_dim_axis=0,
            time_dim_axis=1,
        )



"""
VariableLayer and related
"""

from __future__ import annotations
from typing import Optional, Sequence, List, TypeVar
import contextlib
import tensorflow as tf
import returnn.tf.compat as tf_compat
import returnn.tf.util.basic as tf_util
from returnn.tensor import Tensor, Dim
from .base import LayerBase


T = TypeVar("T")


class VariableLayer(LayerBase):
    """
    Represents a variable. Can add batch/time dimension if wanted. Can be trainable.
    See defaults.
    """

    layer_class = "variable"

    def __init__(
        self,
        shape,
        dtype="float32",
        add_batch_axis=False,
        add_time_axis=False,
        trainable=True,
        non_critical_for_restore=False,
        init=None,
        init_by_layer=None,
        param_name=None,
        **kwargs,
    ):
        """
        :param tuple[int|Dim]|list[int|Dim] shape:
        :param str dtype:
        :param bool add_batch_axis:
        :param bool add_time_axis:
        :param bool trainable:
        :param bool non_critical_for_restore:
        :param str|float|int|None init: see :func:`returnn.tf.util.basic.get_initializer`. 0 by default.
          Alternatively, you can also use option `init_by_layer`.
        :param LayerBase|None init_by_layer:
        :param str|None param_name: self.name (layer name) by default
        """
        shape  # noqa  # used in get_out_data_from_opts
        super(VariableLayer, self).__init__(trainable=trainable, **kwargs)
        assert not self.sources, "%s: does not expect any sources" % self
        self.init_by_layer = init_by_layer
        dim_tags = list(self.output.dim_tags)
        if add_batch_axis:
            assert dim_tags[0].is_batch_dim()
            dim_tags = dim_tags[1:]
        if add_time_axis:
            assert dim_tags[0].dimension == 1
            dim_tags = dim_tags[1:]
        shape_ = [d.dimension for d in dim_tags]
        assert all(shape_), self.output  # all static
        with self.var_creation_scope():
            if init_by_layer is None:
                if init is None:
                    init = 0
                initializer = tf_util.get_initializer(
                    init, dtype=dtype, seed=self.network.random.randint(2**31), eval_local_ns={"layer": self}
                )
            else:
                assert init_by_layer is not None
                out_data_base = Tensor(name=self.output.name, dim_tags=dim_tags, dtype=dtype)
                initializer = init_by_layer.output.copy_compatible_to(out_data_base).placeholder
                shape_ = None  # get_variable requires shape to be not defined when the initializer is another tensor
            self.var = self.add_param(
                tf_compat.v1.get_variable(
                    name=param_name or self.name,
                    shape=shape_,
                    dtype=dtype,
                    initializer=initializer,
                    trainable=trainable,
                ),
                axes_split_info=[d.axis_split_info() for d in dim_tags],
                non_critical_for_restore=non_critical_for_restore,
            )
            out = self.var
            if add_time_axis:
                out = tf.expand_dims(out, axis=0)
            if add_batch_axis:
                # Unbroadcast to not confuse some other layers
                batch_dim = self.output.get_batch_dim()
                out = tf_util.expand_dims_unbroadcast(out, axis=0, dim=batch_dim)
        self.output.placeholder = out

    def get_dep_layers(self):
        """
        :rtype: list[LayerBase]
        """
        deps = super(VariableLayer, self).get_dep_layers()
        if self.init_by_layer:
            deps.append(self.init_by_layer)
        return deps

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """
        :param dict[str] d: will modify inplace
        :param returnn.tf.network.TFNetwork network:
        :param ((str) -> LayerBase) get_layer: function to get or construct another layer
        """
        # Overwrite default behavior for default sources.
        # Here: none by default.
        d.setdefault("from", [])
        super(VariableLayer, cls).transform_config_dict(d, network=network, get_layer=get_layer)
        if d.get("init_by_layer", None):
            d["init_by_layer"] = get_layer(d["init_by_layer"])

    @classmethod
    def get_out_data_from_opts(
        cls, name, network, shape, dtype="float32", add_batch_axis=False, add_time_axis=False, **kwargs
    ):
        """
        :param str name:
        :param returnn.tf.network.TFNetwork network:
        :param tuple[int|Dim]|list[int|Dim] shape:
        :param str dtype:
        :param bool add_batch_axis:
        :param bool add_time_axis:
        :rtype: Tensor
        """
        assert isinstance(shape, (list, tuple))
        assert len(shape) == 0 or all(shape)
        dim_tags = []
        for i, d in enumerate(shape):
            if isinstance(d, Dim):
                assert d.dimension is not None, "%r: need static dims but got %r" % (name, d)
            elif isinstance(d, int):
                d = Dim(
                    kind=Dim.Types.Spatial if i < len(shape) - 1 else Dim.Types.Feature,
                    description="%s:static:%i" % (name, i),
                    auto_generated=True,
                    dimension=d,
                )
            else:
                raise TypeError("Layer %r: invalid type %s in shape %r" % (name, type(d), shape))
            dim_tags.append(d)
        if add_time_axis:
            dim_tags.insert(
                0, Dim(kind=Dim.Types.Time, description="%s:dummy-time" % name, dimension=1, auto_generated=True)
            )
        if add_batch_axis:
            dim_tags.insert(
                0, Dim(kind=Dim.Types.Batch, description="batch", batch=network.get_global_batch_info(), dimension=None)
            )
        return Tensor(
            name="%s_output" % name,
            dim_tags=dim_tags,
            dtype=dtype,
            batch=network.get_global_batch_info() if add_batch_axis else None,
        )


class VariableAssignLayer(LayerBase):
    """
    Assigns a new value to a variable.
    """

    layer_class = "variable_assign"

    def __init__(
        self,
        var: LayerBase,
        value: LayerBase,
        control_dependencies: Optional[Sequence[LayerBase]] = None,
        op: str = "assign",
        **kwargs,
    ):
        """
        :param var:
        :param value:
        :param control_dependencies:
        :param op: "assign" or "add"
        """
        super().__init__(**kwargs)
        self.var = var
        self.value = value
        self.control_dependencies = list(control_dependencies) if control_dependencies else []
        deps = [src.output.placeholder.op for src in self.control_dependencies]

        while not isinstance(var, VariableLayer):
            if isinstance(var, VariableAssignLayer):
                deps.append(var.output.placeholder.op)
                var = var.var
            elif isinstance(var, VariableReadLayer):
                deps.append(var.output.placeholder.op)
                var = var.var
            else:
                raise TypeError(f"{self}: invalid var {var!r}")
        assert isinstance(var, VariableLayer), f"{self}: var must be a VariableLayer, got {var}"
        self.tf_var: tf.Variable = var.var
        assert isinstance(self.tf_var, tf.Variable), f"{self}: var must be a tf.Variable, got {self.tf_var}"

        value_data = value.output.copy_compatible_to(self.var.output)
        with tf.control_dependencies(deps) if deps else contextlib.nullcontext():
            if op == "assign":
                op_ = self.tf_var.assign(value_data.placeholder, read_value=False)
            elif op == "add":
                op_ = self.tf_var.assign_add(value_data.placeholder, read_value=False)
            else:
                raise ValueError(f"{self}: invalid op {op!r}")
        # op_ is only defined in graph-mode. in eager-mode, it's not relevant.
        with tf.control_dependencies([op_]) if op_ is not None else contextlib.nullcontext():
            self.output.placeholder = tf.zeros((), dtype="int32")

    def get_dep_layers(self) -> List[LayerBase]:
        """deps"""
        return super().get_dep_layers() + [self.var, self.value] + self.control_dependencies

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """transform"""
        d.setdefault("from", [])
        super().transform_config_dict(d, network=network, get_layer=get_layer)
        d["var"] = get_layer(d["var"])
        d["value"] = get_layer(d["value"])
        if d.get("control_dependencies"):
            d["control_dependencies"] = [get_layer(layer) for layer in d["control_dependencies"]]

    @classmethod
    def get_out_data_from_opts(cls, name: str, var: LayerBase, **kwargs):
        """out"""
        return Tensor(name, dims=(), dtype="int32")  # dummy, will be just the op


class VariableReadLayer(LayerBase):
    """
    Read a variable (currently expected from VariableLayer).
    Supports control dependencies to exactly specify when it should be read.
    """

    layer_class = "variable_read"

    def __init__(self, var: LayerBase, control_dependencies: Optional[Sequence[LayerBase]] = None, **kwargs):
        """
        :param var: e.g. VariableLayer
        :param control_dependencies: to control what ops must run before the var is read (e.g. assign ops)
        """
        super().__init__(**kwargs)
        self.var = var
        self.control_dependencies = list(control_dependencies) if control_dependencies else []
        deps = [src.output.placeholder.op for src in self.control_dependencies]

        while not isinstance(var, VariableLayer):
            if isinstance(var, VariableAssignLayer):
                deps.append(var.output.placeholder.op)
                var = var.var
            elif isinstance(var, VariableReadLayer):
                deps.append(var.output.placeholder.op)
                var = var.var
            else:
                raise TypeError(f"{self}: invalid var {var!r}")
        assert isinstance(var, VariableLayer), f"{self}: var must be a VariableLayer, got {var}"
        self.tf_var: tf.Variable = var.var
        assert isinstance(self.tf_var, tf.Variable), f"{self}: var must be a tf.Variable, got {self.tf_var}"
        with tf.control_dependencies(deps) if deps else contextlib.nullcontext():
            self.output.placeholder = self.tf_var.read_value()

    def get_dep_layers(self) -> List[LayerBase]:
        """deps"""
        return super().get_dep_layers() + [self.var] + self.control_dependencies

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """transform"""
        d.setdefault("from", [])
        super().transform_config_dict(d, network=network, get_layer=get_layer)
        d["var"] = get_layer(d["var"])
        if d.get("control_dependencies"):
            d["control_dependencies"] = [get_layer(layer) for layer in d["control_dependencies"]]

    @classmethod
    def get_out_data_from_opts(cls, name: str, var: LayerBase, **kwargs):
        """out"""
        return var.output.copy_template(name="%s_output" % name)
